@@ -32,6 +32,7 @@ _sdpads_layout = [
         ("o",  4),
         ("oe", 1)
     ]),
+    ("data_i_ce", 1),
 ]
 
 # SDCard PHY Clocker -------------------------------------------------------------------------------
@@ -39,27 +40,37 @@ _sdpads_layout = [
 class SDPHYClocker(Module, AutoCSR):
     def __init__(self):
         self.divider = CSRStorage(9, reset=256)
-        self.stop    = Signal()
-        self.clk     = Signal()
-        self.ce      = Signal()
+        self.stop    = Signal()        # Stop input (for speed handling/backpressure).
+        self.ce      = Signal()        # CE output  (for logic running in sys_clk domain).
+        self.clk_en  = Signal(reset=1) # Clk enable input (from logic running in sys_clk domain).
+        self.clk     = Signal()        # Clk output (for SDCard pads).
 
         # # #
 
+        # Generate divided versions of sys_clk that will be used as SDCard clk.
         clks = Signal(9)
         self.sync += If(~self.stop, clks.eq(clks + 1))
 
+        # Generate delayed version of the SDCard clk (to do specific actions on change).
         clk   = Signal()
         clk_d = Signal()
         self.sync += clk_d.eq(clk)
 
+        # Select SDCard clk based on divider CSR value.
         cases = {}
         cases["default"] = clk.eq(clks[0])
         for i in range(2, 9):
             cases[2**i] = clk.eq(clks[i-1])
         self.comb += Case(self.divider.storage, cases)
-        self.sync += self.clk.eq(clk)
         self.comb += self.ce.eq(clk & ~clk_d)
 
+        # Ensure we don't get short pulses on the SDCard Clk.
+        ce_delayed = Signal()
+        ce_latched = Signal()
+        self.sync += If(clk_d, ce_delayed.eq(self.clk_en))
+        self.comb += If(clk_d, ce_latched.eq(self.clk_en)).Else(ce_latched.eq(ce_delayed))
+        self.comb += self.clk.eq(~clk & ce_latched)
+
 # SDCard PHY Read ----------------------------------------------------------------------------------
 
 @ResetInserter()
@@ -129,7 +140,7 @@ class SDPHYCMDW(Module):
     def __init__(self):
         self.pads_in  = pads_in  = stream.Endpoint(_sdpads_layout)
         self.pads_out = pads_out = stream.Endpoint(_sdpads_layout)
-        self.sink     = sink     = stream.Endpoint([("data", 8)])
+        self.sink     = sink     = stream.Endpoint([("data", 8), ("cmd_type", 2)])
 
         self.done = Signal()
 
@@ -153,7 +164,7 @@ class SDPHYCMDW(Module):
             If(pads_out.ready,
                 NextValue(count, count + 1),
                 If(count == (8-1),
-                    If(sink.last,
+                    If(sink.last & (sink.cmd_type == SDCARD_CTRL_RESPONSE_NONE),
                         NextState("CLK8")
                     ).Else(
                         sink.ready.eq(1),
@@ -178,38 +189,46 @@ class SDPHYCMDW(Module):
 # SDCard PHY Command Read --------------------------------------------------------------------------
 
 class SDPHYCMDR(Module):
-    def __init__(self, sys_clk_freq, cmd_timeout, cmdw):
+    def __init__(self, sys_clk_freq, cmd_timeout, cmdw, busy_timeout=1):
         self.pads_in  = pads_in  = stream.Endpoint(_sdpads_layout)
         self.pads_out = pads_out = stream.Endpoint(_sdpads_layout)
-        self.sink     = sink     = stream.Endpoint([("length", 8)])
+        self.sink     = sink     = stream.Endpoint([("cmd_type", 2), ("data_type", 2), ("length", 8)])
         self.source   = source   = stream.Endpoint([("data", 8), ("status", 3)])
 
         # # #
 
         timeout = Signal(32, reset=int(cmd_timeout*sys_clk_freq))
         count   = Signal(8)
+        busy    = Signal()
 
         cmdr = SDPHYR(cmd=True, data_width=1, skip_start_bit=False)
         self.comb += pads_in.connect(cmdr.pads_in)
         fsm  = FSM(reset_state="IDLE")
         self.submodules += cmdr, fsm
         fsm.act("IDLE",
-            NextValue(count,   0),
-            NextValue(timeout, timeout.reset),
+            # Preload Timeout with Cmd Timeout.
+            NextValue(timeout, int(cmd_timeout*sys_clk_freq)),
+            # Reset Count/Busy flags.
+            NextValue(count, 0),
+            NextValue(busy, 1),
+            # When the Cmd has been sent to the SDCard, get the response.
             If(sink.valid & pads_out.ready & cmdw.done,
                 NextValue(cmdr.reset, 1),
                 NextState("WAIT"),
             )
         )
         fsm.act("WAIT",
+            # Drive Clk.
             pads_out.clk.eq(1),
+            # Reset CMDR.
             NextValue(cmdr.reset, 0),
+            # Change state on Cmd response start.
             If(cmdr.source.valid,
                 NextState("CMD")
             ),
+            # Timeout.
             NextValue(timeout, timeout - 1),
             If(timeout == 0,
-                sink.ready.eq(1),
                 NextState("TIMEOUT")
             )
         )
@@ -219,12 +238,18 @@ class SDPHYCMDR(Module):
             source.status.eq(SDCARD_STREAM_STATUS_OK),
             source.last.eq(count == (sink.length - 1)),
             source.data.eq(cmdr.source.data),
-            If(source.valid & source.ready,
+            If(cmdr.source.valid & source.ready,
                 cmdr.source.ready.eq(1),
                 NextValue(count, count + 1),
                 If(source.last,
                     sink.ready.eq(1),
-                    If(sink.last,
+                    If(sink.cmd_type == SDCARD_CTRL_RESPONSE_SHORT_BUSY,
+                        # Generate the last valid cycle in BUSY state.
+                        source.valid.eq(0),
+                        # Preload Timeout with Busy Timeout.
+                        NextValue(timeout, int(busy_timeout*sys_clk_freq)),
+                        NextState("BUSY")
+                    ).Elif(sink.data_type == SDCARD_CTRL_DATA_TRANSFER_NONE,
                         NextValue(count, 0),
                         NextState("CLK8")
                     ).Else(
@@ -232,12 +257,34 @@ class SDPHYCMDR(Module):
                     )
                 )
             ),
+            # Timeout.
             NextValue(timeout, timeout - 1),
             If(timeout == 0,
-                sink.ready.eq(1),
                 NextState("TIMEOUT")
             ),
         )
+        fsm.act("BUSY",
+            pads_out.clk.eq(1),
+            # D0 is kept low by the SDCard while busy.
+            If(pads_in.valid & pads_in.data.i[0],
+                NextValue(busy, 0),
+            ),
+            # Generate the last cycle of the Cmd Response when no longer busy.
+            If(~busy,
+                source.valid.eq(1),
+                source.last.eq(1),
+                source.status.eq(SDCARD_STREAM_STATUS_OK),
+                If(source.ready,
+                    NextValue(count, 0),
+                    NextState("CLK8")
+                )
+            ),
+            # Timeout.
+            NextValue(timeout, timeout - 1),
+            If(timeout == 0,
+                NextState("TIMEOUT")
+            )
+        )
         fsm.act("CLK8",
             pads_out.clk.eq(1),
             pads_out.cmd.oe.eq(1),
@@ -250,10 +297,11 @@ class SDPHYCMDR(Module):
             )
         )
         fsm.act("TIMEOUT",
+            sink.ready.eq(1), # Ack Cmd.
             source.valid.eq(1),
-            source.status.eq(SDCARD_STREAM_STATUS_TIMEOUT),
             source.last.eq(1),
-            If(source.valid & source.ready,
+            source.status.eq(SDCARD_STREAM_STATUS_TIMEOUT),
+            If(source.ready,
                 NextState("IDLE")
             )
         )
@@ -387,12 +435,10 @@ class SDPHYDATAR(Module):
             NextValue(count, 0),
             If(sink.valid & pads_out.ready,
                 pads_out.clk.eq(1),
-                If(pads_out.ready,
-                    NextValue(timeout, timeout.reset),
-                    NextValue(count, 0),
-                    NextValue(datar.reset, 1),
-                    NextState("WAIT")
-                )
+                NextValue(timeout, timeout.reset),
+                NextValue(count, 0),
+                NextValue(datar.reset, 1),
+                NextState("WAIT")
             )
         )
         fsm.act("WAIT",
@@ -451,29 +497,39 @@ class SDPHYDATAR(Module):
             source.valid.eq(1),
             source.status.eq(SDCARD_STREAM_STATUS_TIMEOUT),
             source.last.eq(1),
-            If(source.valid & source.ready,
+            If(source.ready,
                 NextState("IDLE")
             )
         )
 
 # SDCard PHY IO ------------------------------------------------------------------------------------
 
-class SDPHYIOGen(Module):
+class SDPHYIO(Module):
+    def __init__(self, clocker, sdpads, round_trip_latency=2):
+        # Generate a data_i_ce pulse round_trip_latency cycles after clocker.clk goes high so that
+        # the data input effectively get sampled on the first sys_clk after the SDCard clk goes high.
+        clocker_clk_delay = Signal(round_trip_latency)
+        self.sync += clocker_clk_delay.eq(Cat(clocker.clk, clocker_clk_delay))
+        self.sync += sdpads.data_i_ce.eq(clocker_clk_delay[-1] & ~clocker_clk_delay[-2])
+
+
+class SDPHYIOGen(SDPHYIO):
     def __init__(self, clocker, sdpads, pads):
+        SDPHYIO.__init__(self, clocker, sdpads, round_trip_latency=2)
         # Rst
         if hasattr(pads, "rst"):
             self.comb += pads.rst.eq(0)
 
         # Clk
         self.specials += SDROutput(
-            clk = ClockSignal(),
-            i   = ~clocker.clk & sdpads.clk,
+            clk = ClockSignal("sys"),
+            i   = clocker.clk,
             o   = pads.clk
         )
 
         # Cmd
         self.specials += SDRTristate(
-            clk = ClockSignal(),
+            clk = ClockSignal("sys"),
             io  = pads.cmd,
             o   = sdpads.cmd.o,
             oe  = sdpads.cmd.oe,
@@ -483,7 +539,7 @@ class SDPHYIOGen(Module):
         # Data
         for i in range(4):
             self.specials += SDRTristate(
-                clk = ClockSignal(),
+                clk = ClockSignal("sys"),
                 io  = pads.data[i],
                 o   = sdpads.data.o[i],
                 oe  = sdpads.data.oe,
@@ -494,17 +550,17 @@ class SDPHYIOGen(Module):
         if hasattr(pads, "cmd_dir"):
             self.specials += [
                 SDROutput(
-                    clk = ClockSignal(),
+                    clk = ClockSignal("sys"),
                     i   = sdpads.cmd.oe,
                     o   = pads.cmd_dir,
                 ),
                 SDROutput(
-                    clk = ClockSignal(),
+                    clk = ClockSignal("sys"),
                     i   = sdpads.data.oe,
                     o   = pads.dat0_dir,
                 ),
                 SDROutput(
-                    clk = ClockSignal(),
+                    clk = ClockSignal("sys"),
                     i   = sdpads.data.oe,
                     o   = pads.dat13_dir,
                 )
@@ -512,10 +568,11 @@ class SDPHYIOGen(Module):
 
 # SDCard PHY Emulator ------------------------------------------------------------------------------
 
-class SDPHYIOEmulator(Module):
+class SDPHYIOEmulator(SDPHYIO):
     def __init__(self, clocker, sdpads, pads):
+        SDPHYIO.__init__(self, clocker, sdpads, round_trip_latency=2) # FIXME: check round_trip_latency.
         # Clk
-        self.comb += If(sdpads.clk, pads.clk.eq(~clocker.clk))
+        self.comb += pads.clk.eq(clocker.clk)
 
         # Cmd
         self.comb += [
@@ -567,10 +624,11 @@ class SDPHY(Module, AutoCSR):
         ]
         for m in [init, cmdw, cmdr, dataw, datar]:
             self.comb += m.pads_out.ready.eq(self.clocker.ce)
+        self.comb += self.clocker.clk_en.eq(sdpads.clk)
 
         # Connect physical pads to pads_in of submodules -------------------------------------------
         for m in [init, cmdw, cmdr, dataw, datar]:
-            self.comb += m.pads_in.valid.eq(self.clocker.ce)
+            self.comb += m.pads_in.valid.eq(sdpads.data_i_ce)
             self.comb += m.pads_in.cmd.i.eq(sdpads.cmd.i)
             self.comb += m.pads_in.data.i.eq(sdpads.data.i)
 
